@@ -20,4 +20,4 @@ extern char *copystring(char *, const char *);
 
 extern scheme_value strlen_or_false(const char *);
 
-extern void cig_check_nargs(int arity, int nargs, char *fn);
+extern void cig_check_nargs(int arity, int nargs, const char *fn);
