@@ -384,6 +384,7 @@
 	  enabled-interrupts
 	  one-second
 	  all-interrupts
+	  with-interrupts		; For scsh
 	  with-interrupts-allowed
 	  with-interrupts-inhibited
 	  interrupt-before-heap-overflow!
@@ -731,11 +732,15 @@
 (define-interface scan-interface
   (export scan-forms
 	  scan-file
+	  really-scan-file
+	  really-scan-port		; for scsh
 	  scan-structures		;load-package.scm, link/link.scm
+	  really-scan-structures
 	  scan-package
 	  set-optimizer!
 	  $note-file-package
 	  noting-undefined-variables
+	  really-noting-undefined-variables
 	  $note-undefined		;env/command.scm
 	  note-undefined!
 	  check-structure))
@@ -759,6 +764,7 @@
 (define-interface compiler-interface
   (export compile
 	  compile-and-run-file		;for LOAD
+	  compile-and-run-port		;for LOAD (in scsh)
 	  compile-and-run-forms		;for EVAL
 	  compile-and-run-scanned-forms ;for eval.scm / ensure-loaded
 	  compile-file			;link/link.scm
@@ -799,7 +805,8 @@
 	  with-fresh-compiler-state))  ;for linker
 
 (define-interface evaluation-interface
-  (export eval load load-into eval-from-file eval-scanned-forms))
+  (export eval load load-into load-quietly eval-from-file
+	  eval-scanned-forms really-eval-scanned-forms))
 
 (define-interface environments-interface  ;cheesy
   (export *structure-ref
