@@ -2,14 +2,9 @@
 ;;; Copyright (c) 1994 by Olin Shivers.
 ;;; Copyright (c) 1994 by Brian D. Carlstrom.
 
-(define-syntax define-signals
-  (syntax-rules ()
-    ((define-signals form ...)
-     (begin (define-enum-constant "signal" . form) ...))))
-
 ;;POSIX only defined here.
 
-(define-signals
+(define-enum-constants signal
   ;; POSIX
   (hup	1)	; hangup 
   (int	2)	; interrupt 
@@ -33,3 +28,6 @@
   (usr1 30)	; user defined signal 1 
   (usr2 31)	; user defined signal 2 
   )
+
+(define signals-ignored-by-default
+  (list signal/chld signal/cont))			; These are Posix.
