@@ -1,16 +1,15 @@
 ;;; Networking for the Scheme Shell
-;;; Copyright (c) 1994 by Brian D. Carlstrom.
+;;; Copyright (c) 1994-1995 by Brian D. Carlstrom.
 ;;; Copyright (c) 1994 by Olin Shivers.
 
 ;;; Scheme48 implementation.
 
-;;; TODO
-;;; high level interface
-
 (foreign-source
  "#include <sys/types.h>"
  "#include <sys/socket.h>"
  ""
+ "/* Make sure foreign-function stubs interface to the C funs correctly: */"
+ "#include \"network1.h\""
  ""
  "extern int errno;"
  "extern int h_errno;"
@@ -38,19 +37,21 @@
 					  (service-info:port 
 					   (service-info (cadr args) "tcp")))
 					 (else
-					  (error "socket-connect: bad arg")))))
+					  (error
+					   "socket-connect: bad arg ~s"
+					   args)))))
 			(internet-address->socket-address host port)))
 		     ((= protocol-family
 			 protocol-family/unix)
 		      (unix-address->socket-address (car args)))
 		     (else 
-		      (error "socket-connect: unsupported protocol-family")))))
+		      (error "socket-connect: unsupported protocol-family ~s"
+			     protocol-family)))))
     (connect-socket sock addr)
     sock))
 
 (define (bind-listen-accept-loop protocol-family proc arg)
-  (let* ((sock (create-socket protocol-family/internet 	
-			      socket-type/stream))
+  (let* ((sock (create-socket protocol-family socket-type/stream))
 	 (addr (cond ((= protocol-family
 			 protocol-family/internet)
 		      (let ((port (cond ((integer? arg) arg)
@@ -58,14 +59,16 @@
 					 (service-info:port 
 					  (service-info arg "tcp")))
 					(else
-					 (error "socket-connect: bad arg")))))
+					 (error "socket-connect: bad arg ~s"
+						arg)))))
 			(internet-address->socket-address internet-address/any
 							  arg)))
 		     ((= protocol-family
 			 protocol-family/unix)
 		      (unix-address->socket-address arg))
 		     (else 
-      (error "bind-listen-accept-loop: unsupported protocol-family")))))
+      (error "bind-listen-accept-loop: unsupported protocol-family ~s"
+	     protocol-family)))))
     (set-socket-option sock level/socket socket/reuse-address #t)
     (bind-socket sock addr)
     (listen-socket sock 5)
@@ -97,9 +100,11 @@
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (internet-address->socket-address address32 port16)
   (cond ((not (<= 0 address32 #xffffffff))
-	 (error "internet-address->socket-address: address out of range"))
+	 (error "internet-address->socket-address: address out of range ~s"
+		address32))
 	((not (<= 0 port16 #xffff))
-	 (error "internet-address->socket-address: port out of range"))
+	 (error "internet-address->socket-address: port out of range ~s"
+		port16))
 	(else 
 	 (make-socket-address address-family/internet
 			      (string-append (integer->string address32) 
@@ -109,7 +114,8 @@
   (if (or (not (socket-address? sockaddr))
 	  (not (= (socket-address:family sockaddr) 
 		  address-family/internet)))
-      (error "socket-address->internet-address: internet socket expected")
+      (error "socket-address->internet-address: internet socket expected ~s"
+	     sockaddr)
       (values (string->integer (substring (socket-address:address sockaddr) 
 					  0 4))
 	      (string->integer (substring (socket-address:address sockaddr)
@@ -117,35 +123,35 @@
 
 (define (unix-address->socket-address path)
   (if (> (string-length path) 108)
-      (error "unix-address->socket-address: path too long")
+      (error "unix-address->socket-address: path too long ~s" path)
       (make-socket-address address-family/unix path)))
 
 (define (socket-address->unix-address sockaddr)
   (if (or (not (socket-address? sockaddr))
 	  (not (= (socket-address:family sockaddr) 
 		  address-family/unix)))
-      (error "socket-address->unix-address expects an unix socket")
-      socket-address:address))
+      (error "socket-address->unix-address expects an unix socket ~s" sockaddr)
+      (socket-address:address sockaddr)))
 
 (define (make-addr af)
   (make-string (cond ((= af address-family/unix) 108)
 		     ((= af address-family/internet) 8)
-		     (else (error "make-addr: unknown address-family")))))
+		     (else 
+		      (error "make-addr: unknown address-family ~s" af)))))
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 ;;; socket syscall
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (create-socket pf type . maybe-protocol)
-  (receive (protocol)
-	   (parse-optionals maybe-protocol 0)
+  (let ((protocol (:optional maybe-protocol 0)))
     (if (not (and (integer? pf)
 		  (integer? type)
 		  (integer? protocol)))
-	(error "create-socket: integer arguments expected")
+	(error "create-socket: integer arguments expected ~s ~s ~s" 
+	       pf type protocol)
 	(let* ((fd  (%socket pf type protocol))
-	       (in  (make-input-fdport fd))
+	       (in  (make-input-fdport fd 0))
 	       (out (dup->outport in)))
-	  (%install-port fd in)
 	  (make-socket pf in out)))))
 
 (define-foreign %socket/errno
@@ -170,14 +176,15 @@
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (bind-socket sock name)
   (cond ((not (socket? sock))
-	 (error "bind-socket: socket expected"))
+	 (error "bind-socket: socket expected ~s" sock))
 	((not (socket-address? name))
-	 (error "bind-socket: socket-address expected"))
+	 (error "bind-socket: socket-address expected ~s" name))
 	(else
 	 (let ((family (socket:family sock)))
 	   (if (not (= family (socket-address:family name)))
 	       (error 
-		"bind-socket: trying to bind incompatible address to socket")
+		"bind-socket: trying to bind incompatible address to socket ~s"
+		name)
 	       (%bind (socket->fdes sock)
 		      family
 		      (socket-address:address name)))))))
@@ -188,21 +195,22 @@
 	       (string-desc name))	; scheme descriptor
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (%bind sockfd family name) %bind/errno)
+(define-errno-syscall (%bind sockfd family name) %bind/errno)
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 ;;; connect syscall
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (connect-socket sock name)
   (cond ((not (socket? sock))
-	 (error "connect-socket: socket expected"))
+	 (error "connect-socket: socket expected ~s" sock))
 	((not (socket-address? name))
-	 (error "connect-socket: socket-address expected"))
+	 (error "connect-socket: socket-address expected ~s" name))
 	(else
 	 (let ((family (socket:family sock)))
 	   (cond ((not (= family (socket-address:family name)))
 		  (error 
-	   "connect: trying to connect socket to incompatible address"))
+	   "connect: trying to connect socket to incompatible address ~s"
+	   name))
 		 (else
 		  (%connect (socket->fdes sock)
 			    (socket:family sock)
@@ -214,16 +222,16 @@
 		  (desc    name))	; scheme descriptor
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (%connect sockfd family name) %connect/errno)
+(define-errno-syscall (%connect sockfd family name) %connect/errno)
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 ;;; listen syscall
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (listen-socket sock backlog)
   (cond ((not (socket? sock))
-	 (error "listen-socket: socket expected"))
+	 (error "listen-socket: socket expected ~s" sock))
 	((not (integer? backlog))
-	 (error "listen-socket: integer expected"))
+	 (error "listen-socket: integer expected ~s" backlog))
 	(else
 	 (%listen (socket->fdes sock) backlog))))
 	 
@@ -232,20 +240,19 @@
 	  (integer backlog))	; backlog
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (%listen sockfd backlog) %listen/errno)
+(define-errno-syscall (%listen sockfd backlog) %listen/errno)
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 ;;; accept syscall
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (accept-connection sock)
   (if (not (socket? sock))
-      (error "accept-connection: socket expected")
+      (error "accept-connection: socket expected ~s" sock)
       (let* ((family (socket:family sock))
 	     (name   (make-addr family))
 	     (fd     (%accept (socket->fdes sock) family name))
-	     (in     (make-input-fdport fd))
+	     (in     (make-input-fdport fd 0))
 	     (out    (dup->outport in)))
-	(%install-port fd in)
 	(values (make-socket family in out)
 		(make-socket-address family name)))))
 
@@ -265,7 +272,7 @@
 (define (socket-remote-address sock)
   (if (or (not (socket? sock))
 	  (not (= (socket:family sock) address-family/internet)))
-      (error "socket-remote-address: internet socket expected")
+      (error "socket-remote-address: internet socket expected ~s" sock)
       (let* ((family (socket:family sock))
 	     (name   (make-addr family)))
 	(%peer-name (socket->fdes sock)
@@ -279,7 +286,7 @@
 		    (string-desc name))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (%peer-name sock family name) %peer-name/errno)
+(define-errno-syscall (%peer-name sock family name) %peer-name/errno)
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 ;;; getsockname syscall
@@ -287,7 +294,7 @@
 (define (socket-local-address sock)
   (if (or (not (socket? sock))
 	  (not (= (socket:family sock) address-family/internet)))
-      (error "socket-local-address: internet socket expected")
+      (error "socket-local-address: internet socket expected ~s" sock)
       (let* ((family (socket:family sock))
 	     (name   (make-addr family)))
 	(%socket-name (socket->fdes sock)
@@ -301,7 +308,7 @@
 		      (string-desc name))
   (to-scheme integer "False_on_zero"))
 
-(define-simple-errno-syscall 
+(define-errno-syscall 
   (%socket-name sock family name) %socket-name/errno)
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
@@ -309,9 +316,9 @@
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (shutdown-socket sock how)
   (cond ((not (socket? sock))
-	 (error "shutdown-socket: socket expected"))
+	 (error "shutdown-socket: socket expected ~s" sock))
 	((not (integer? how))
-	 (error "shutdown-socket: integer expected"))
+	 (error "shutdown-socket: integer expected ~s" how))
 	(else
 	 (%shutdown (socket->fdes sock) how))))
 
@@ -320,7 +327,7 @@
 	    (integer how))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall 
+(define-errno-syscall 
   (%shutdown sock how) %shutdown/errno)
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
@@ -328,15 +335,13 @@
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (create-socket-pair type)
   (if (not (integer? type))
-      (error "create-socket-pair: integer argument expected")
+      (error "create-socket-pair: integer argument expected ~s" type)
       (receive (s1 s2)
 	       (%socket-pair type)
-        (let* ((in1  (make-input-fdport s1))
+        (let* ((in1  (make-input-fdport s1 0))
 	       (out1 (dup->outport in1))
-	       (in2  (make-input-fdport s2))
+	       (in2  (make-input-fdport s2 0))
 	       (out2 (dup->outport in2)))
-	  (%install-port s1 in1)
-	  (%install-port s2 in2)
 	  (values (make-socket protocol-family/unix in1 out1)
 		  (make-socket protocol-family/unix in2 out2))))))
 
@@ -356,39 +361,38 @@
 ;;; recv syscall
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 (define (receive-message socket len . maybe-flags) 
-  (receive (flags)
-	   (parse-optionals maybe-flags 0)
+  (let ((flags (:optional maybe-flags 0)))
     (cond ((not (socket? socket))
-	   (error "receive-message: socket expected"))
+	   (error "receive-message: socket expected ~s" socket))
 	  ((or (not (integer? flags))
 	       (not (integer? len)))
-	   (error "receive-message: integer expected"))
+	   (error "receive-message: integer expected ~s ~s" flags len))
 	  (else 
 	   (let ((s (make-string len)))
 	     (receive (nread from)
-		      (receive-message! socket flags s 0 len)
+		      (receive-message! socket s 0 len flags)
                (values
 		(cond ((not nread) #f)	; EOF
 		      ((= nread len) s)
 		      (else (substring s 0 nread)))
 		from)))))))
 
-(define (receive-message! socket s . maybe-args)
-  (receive (start end flags)
-	   (parse-optionals maybe-args 0 (string-length s) 0)
-    (cond ((not (socket? socket))
-	   (error "receive-message!: socket expected"))
-	  ((not (or (integer? flags)
-		    (integer? start)
-		    (integer? end)))
-	   (error "receive-message!: integer expected"))
-	  ((not (string? s))
-	   (error "receive-message!: string expected"))
-	  (else 
-           (generic-receive-message! (socket->fdes socket) flags
-				 s start end 
-				 recv-substring!/errno
-				 (socket:family socket))))))
+(define (receive-message! socket s . args)
+  (if (not (string? s))
+      (error "receive-message!: string expected ~s" s)
+      (let-optionals args ((start 0) (end (string-length s)) (flags 0))
+        (cond ((not (socket? socket))
+	       (error "receive-message!: socket expected ~s" socket))
+	      ((not (or (integer? flags)
+			(integer? start)
+			(integer? end)))
+	       (error "receive-message!: integer expected ~s ~s ~s"
+		      flags start end))
+	      (else 
+	       (generic-receive-message! (socket->fdes socket) flags
+					 s start end 
+					 recv-substring!/errno
+					 (socket:family socket)))))))
 
 (define (generic-receive-message! sockfd flags s start end reader from)
   (if (bogus-substring-spec? s start end)
@@ -413,34 +417,38 @@
 		   (else (loop (+ i nread)))))))))
 
 (define (receive-message/partial socket len . maybe-flags)
-  (receive (flags)
-	   (parse-optionals maybe-flags 0)
+  (let ((flags (:optional maybe-flags 0)))
     (cond ((not (socket? socket))
-	   (error "receive-message/partial: socket expected"))
+	   (error "receive-message/partial: socket expected ~s" socket))
 	  ((or (not (integer? flags))
 	       (not (integer? len)))
-	   (error "receive-message/partial: integer expected"))
+	   (error "receive-message/partial: integer expected ~s ~s" flags len))
 	  (else 
-	   (let* ((s (make-string len))
-		  (nread (receive-message!/partial socket flags s 0 len )))
-	     (cond ((not nread) #f)	; EOF
-		   ((= nread len) s)
-		   (else (substring s 0 nread))))))))
+	   (let ((s (make-string len)))
+	     (receive (nread addr)
+		      (receive-message!/partial socket s 0 len flags)
+		      (values 
+		       (cond ((not nread) #f)	; EOF
+			     ((= nread len) s)
+			     (else (substring s 0 nread)))
+		       addr)))))))
 
-(define (receive-message!/partial socket s . maybe-args)
-  (receive (start end flags)
-	   (parse-optionals maybe-args 0 (string-length s) 0)
-    (cond ((not (socket? socket))
-	   (error "receive-message!/partial: socket expected"))
-	  ((not (integer? flags))
-	   (error "receive-message!/partial: integer expected"))
-	  ((not (string? s))
-	   (error "receive-message!/partial: string expected"))
-	  (else 
-	   (generic-receive-message!/partial (socket->fdes socket) flags 
-					     s start end
-					     recv-substring!/errno
-					     (socket:family socket))))))
+(define (receive-message!/partial socket s . args)
+  (if (not (string? s))
+      (error "receive-message!/partial: string expected ~s" s)
+      (let-optionals args ((start 0) (end (string-length s)) (flags 0))
+        (cond ((not (socket? socket))
+	       (error "receive-message!/partial: socket expected ~s"
+		      socket))
+	      ((not (integer? flags))
+	       (error "receive-message!/partial: integer expected ~s"
+		      flags))
+	      (else 
+	       (generic-receive-message!/partial (socket->fdes socket)
+						 flags 
+						 s start end
+						 recv-substring!/errno
+						 (socket:family socket)))))))
 
 (define (generic-receive-message!/partial sockfd flags s start end reader from)
   (if (bogus-substring-spec? s start end)
@@ -451,16 +459,20 @@
 	(let loop ()
 	  (receive (err nread) 
 		   (reader sockfd flags s start end addr)
-	     (if err
-		 (case err
-		   ((errno/intr) (loop))
+
+	     (cond ((not err)
+		    (values (and (not (zero? nread)) nread)
+			    (make-socket-address from addr)))
+
+		   ((= err errno/intr) (loop))
+
 		   ; No forward-progess here.
-		   ((errno/wouldblock errno/again) 0) 
+		   ((or (= err errno/wouldblock)
+			(= err errno/again))
+		    0) 
+
 		   (else (errno-error err reader sockfd flags
-				      s start start end addr)))
-		 (values
-		  (and (not (zero? nread)) nread)
-		  from)))))))
+				      s start start end addr))))))))
 
 (define-foreign recv-substring!/errno
   (recv_substring (integer sockfd)
@@ -476,16 +488,14 @@
 ;;; send syscall
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
 
-(define (send-message socket s . maybe-args)
-  (receive (start end flags addr)
-	   (parse-optionals maybe-args
-			    0 (string-length s) 0 #f)
+(define (send-message socket s . args)
+  (let-optionals args ((start 0) (end (string-length s)) (flags 0) (addr #f))
     (cond ((not (socket? socket))
-	   (error "send-message: socket expected"))
+	   (error "send-message: socket expected ~s" socket))
 	  ((not (integer? flags))
-	   (error "send-message: integer expected"))
+	   (error "send-message: integer expected ~s" flags))
 	  ((not (string? s))
-	   (error "send-message: string expected"))
+	   (error "send-message: string expected ~s" s))
 	  (else 
 	   (generic-send-message (socket->fdes socket) flags
 				 s start end
@@ -503,31 +513,26 @@
       (if (< i end)
 	  (receive (err nwritten) 
 		   (writer sockfd flags s i end family addr)
-	     (if err
-		 (case err
-		   ((errno/intr) (loop i))
-		   (else (errno-error err sockfd flags family addr
-				      s start i end writer)))
-		 (loop (+ i nwritten))))))))
+	    (cond ((not err) (loop (+ i nwritten)))
+		  ((= err errno/intr) (loop i))
+		  (else (errno-error err sockfd flags family addr
+				     s start i end writer))))))))
 
-(define (send-message/partial socket s . maybe-args)
-  (receive (start end flags addr)
-	   (parse-optionals maybe-args
-			    0 (string-length s) 0 #f)
+
+(define (send-message/partial socket s . args)
+  (let-optionals args ((start 0) (end (string-length s)) (flags 0) (addr #f))
     (cond ((not (socket? socket))
-	   (error "send-message/partial: socket expected"))
+	   (error "send-message/partial: socket expected ~s" socket))
 	  ((not (integer? flags))
-	   (error "send-message/partial: integer expected"))
+	   (error "send-message/partial: integer expected ~s" flags))
 	  ((not (string? s))
-	   (error "send-message/partial: string expected"))
+	   (error "send-message/partial: string expected ~s" s))
 	  (else 
            (generic-send-message/partial (socket->fdes socket) flags
 					 s start end
 					 send-substring/errno
 					 (if addr (socket-address:family addr) 0)
-					 (and addr 
-					      (socket-address:family 
-					       addr)))))))
+					 (if addr (socket-address:address addr)))))))
 
 (define (generic-send-message/partial sockfd flags s start end writer family addr)
   (if (bogus-substring-spec? s start end)
@@ -536,17 +541,14 @@
 	     s start end writer))
 
   (if (= start end) 0			; Vacuous request.
-      (let ((addr (if addr (make-addr family) "")))
-	(let loop ()
-	  (receive (err nwritten) 
-		   (writer sockfd flags s start end family addr)
-	     (if err
-		 (case err
-		   ((errno/intr) (loop))
-		   ((errno/again errno/wouldblock) 0)
-		   (else (errno-error err sockfd flags family addr
-				      s start start end writer)))
-		 nwritten))))))
+      (let loop ()
+	(receive (err nwritten) 
+		 (writer sockfd flags s start end family addr)
+		 (cond ((not err) nwritten)
+		       ((= err errno/intr) (loop))
+		       ((or (= err errno/again) (= err errno/wouldblock)) 0)
+		       (else (errno-error err sockfd flags family addr
+					  s start start end writer)))))))
 
 (define-foreign send-substring/errno
   (send_substring (integer sockfd)
@@ -565,29 +567,33 @@
 
 (define (socket-option sock level option)
   (cond ((not (socket? sock))
-	 (error "socket-option: socket expected"))
+	 (error "socket-option: socket expected ~s" sock))
 	((or (not (integer? level))(not (integer? option)))
-	 (error "socket-option: integer expected"))
+	 (error "socket-option: integer expected ~s ~s" level option))
 	((boolean-option? option)
 	 (let ((result (%getsockopt (socket->fdes sock) level option)))
-	   (cond ((= result -1) (error "socket-option"))
+	   (cond ((= result -1) 
+		  (error "socket-option ~s ~s ~s" sock level option))
 		 (else (not (= result 0))))))
 	((value-option? option)
 	 (let ((result (%getsockopt (socket->fdes sock) level option)))
-	   (cond ((= result -1) (error "socket-option"))
+	   (cond ((= result -1) 
+		  (error "socket-option ~s ~s ~s" sock level option))
 		 (else result))))
 	((linger-option? option)
 	 (receive (result/on-off time)
 		  (%getsockopt-linger (socket->fdes sock) level option)
-	    (cond ((= result/on-off -1) (error "socket-option"))
+	    (cond ((= result/on-off -1) 
+		   (error "socket-option ~s ~s ~s" sock level option))
 		  (else (if (= result/on-off 0) #f time)))))
 	((timeout-option? option)
 	 (receive (result/secs usecs)
 		  (%getsockopt-linger (socket->fdes sock) level option)
-	    (cond ((= result/secs -1) (error "socket-option"))
-		  (else (+ result/secs (/ usecs 1000))))))
+	   (cond ((= result/secs -1) 
+		  (error "socket-option ~s ~s ~s" sock level option))
+		 (else (+ result/secs (/ usecs 1000))))))
 	(else
-	 "socket-option: unknown option type")))
+	 "socket-option: unknown option type ~s" option)))
 
 (define-foreign %getsockopt/errno
   (scheme_getsockopt (integer sockfd)
@@ -631,9 +637,9 @@
 
 (define (set-socket-option sock level option value)
   (cond ((not (socket? sock))
-	 (error "set-socket-option: socket expected"))
+	 (error "set-socket-option: socket expected ~s" sock))
 	((or (not (integer? level)) (not (integer? option)))
-	 (error "set-socket-option: integer expected"))
+	 (error "set-socket-option: integer expected ~s ~s" level option))
 	((boolean-option? option)
 	 (%setsockopt (socket->fdes sock) level option (if value 1 0)))
 	((value-option? option)
@@ -658,7 +664,7 @@
 		     (integer optval))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall 
+(define-errno-syscall 
   (%setsockopt sock level option value) %setsockopt/errno)
 
 
@@ -670,7 +676,7 @@
 			    (integer time))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall 
+(define-errno-syscall 
   (%setsockopt-linger sock level option on-off time) %setsockopt-linger/errno)
 
 (define-foreign %setsockopt-timeout/errno
@@ -681,7 +687,7 @@
 			     (integer usecs))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall 
+(define-errno-syscall 
   (%setsockopt-timeout sock level option secs usecs) %setsockopt-timeout/errno)
 
 ;;;-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-
@@ -706,27 +712,30 @@
 (define-record host-info
   name					; Host name
   aliases				; Alternative names
-  addresses)				; Host addresses
+  addresses				; Host addresses
+
+  ((disclose hi)			; Make host-info records print like
+   (list "host" (host-info:name hi))))	; #{host clark.lcs.mit.edu}.
 
 (define (host-info arg)
   (cond ((string? arg) (name->host-info arg))
 	((socket-address? arg) (address->host-info arg))
-	(else (error "host-info: string or socket-address expected"))))
+	(else (error "host-info: string or socket-address expected ~s" arg))))
 
 (define (address->host-info name)
   (if (or (not (socket-address? name)) 
 	  (not (= (socket-address:family name) address-family/internet)))
-      (error "address->host-info: internet address expected")
+      (error "address->host-info: internet address expected ~s" name)
       (receive (herrno name aliases addresses)
 		  (%host-address->host-info/h-errno 
 		   (socket-address:address name))
 	 (if herrno
-	     (error "address->host-info: non-zero herrno" herrno)
+	     (error "address->host-info: non-zero herrno ~s ~s" name herrno)
 	     (make-host-info name 
 			     (vector->list
 			      (C-string-vec->Scheme aliases   #f))
 			     (vector->list
-			      (C-string-vec->Scheme addresses #f)))))))
+			      (C-long-vec->Scheme addresses #f)))))))
 
 (define-foreign %host-address->host-info/h-errno
   (scheme_host_address2host_info (string-desc name))
@@ -737,11 +746,11 @@
   
 (define (name->host-info name)
   (if (not (string? name))
-      (error "name->host-info: string expected")
+      (error "name->host-info: string expected ~s" name)
       (receive (herrno name aliases addresses)
 	       (%host-name->host-info/h-errno name)
 	 (if herrno
-	     (error "name->host-info: non-zero herrno" herrno)
+	     (error "name->host-info: non-zero herrno ~s ~s" herrno name)
 	     (make-host-info name
 			     (vector->list
 			      (C-string-vec->Scheme aliases   #f))
@@ -767,11 +776,12 @@
 (define (network-info arg)
   (cond ((string? arg) (name->network-info arg))
 	((socket-address? arg) (address->network-info arg))
-	(else (error "network-info: string or socket-address expected"))))
+	(else 
+	 (error "network-info: string or socket-address expected ~s" arg))))
 
 (define (address->network-info name)
   (if (not (integer? name))
-      (error "address->network-info: integer expected")
+      (error "address->network-info: integer expected ~s" name)
       (let ((name (integer->string name))
 	    (net (make-string 4)))
 	(receive (result name aliases)
@@ -782,9 +792,7 @@
 			     (string->integer net))))))
 		  
 (define-foreign %net-address->network-info
-  (scheme_net_address2net_info 
-   (string-desc name)
-   (string-desc net))
+  (scheme_net_address2net_info (string-desc name) (string-desc net))
   (to-scheme integer "False_on_zero")
   static-string	; net name
   (C char**))   ; alias list
@@ -792,7 +800,7 @@
   
 (define (name->network-info name)
   (if (not (string? name))
-      (error "name->network-info: string expected")
+      (error "name->network-info: string expected ~s" name)
       (let ((net (make-string 4)))
 	(receive (result name aliases)
 		 (%net-name->network-info name net)
@@ -802,9 +810,7 @@
 			      (string->integer net))))))
 		  
 (define-foreign %net-name->network-info
-  (scheme_net_name2net_info 
-   (string name)
-   (string net))
+  (scheme_net_name2net_info (string name) (string-desc net))
   (to-scheme integer "False_on_zero")
   static-string	 ; net name
   (C char**))    ; alias list
@@ -819,30 +825,27 @@
   protocol)				; Protocol name
 
 (define (service-info . args)
-  (cond ((string? (car args))  (apply name->service-info args))
-	((integer? (car args)) (apply port->service-info args))
-	(else (error "service-info: string or integer expected"))))
+  (apply (cond ((string?  (car args)) name->service-info)
+	       ((integer? (car args)) port->service-info)
+	       (else (error "service-info: string or integer expected ~s" args)))
+	 args))
 
 (define (port->service-info name . maybe-proto)
-  (receive (proto)
-	   (parse-optionals maybe-proto "")
+  (let ((proto (:optional maybe-proto "")))
     (cond ((not (integer? name))
-	   (error "port->service-info: integer expected"))
+	   (error "port->service-info: integer expected ~s" name))
 	  ((not (string? proto))
-	   (error "port->service-info: string expected"))
+	   (error "port->service-info: string expected ~s" proto))
 	  (else
 	   (receive (result name aliases port protocol)
 		    (%service-port->service-info name proto)
-		    (make-service-info name 
-				       (vector->list
-					(C-string-vec->Scheme aliases #f))
-				       port
-				       protocol))))))
+	     (make-service-info name 
+				(vector->list (C-string-vec->Scheme aliases #f))
+				port
+				protocol))))))
 		  
 (define-foreign %service-port->service-info
-  (scheme_serv_port2serv_info 
-   (integer name)
-   (string  proto))
+  (scheme_serv_port2serv_info (integer name) (string  proto))
   (to-scheme integer "False_on_zero")
   static-string	 ; service name
   (C char**)     ; alias list
@@ -851,23 +854,13 @@
   
   
 (define (name->service-info name . maybe-proto)
-  (receive (proto)
-	   (parse-optionals maybe-proto "")
-    (if (or (not (string? name))
-	    (not (string? name)))
-	(error "name->service-info: string expected")
-	(receive (result name aliases port protocol)
-		 (%service-name->service-info name proto)
-	   (make-service-info name
-			      (vector->list
-			       (C-string-vec->Scheme aliases #f))
-			      port
-			      protocol)))))
+  (receive (result name aliases port protocol)
+      (%service-name->service-info name (:optional maybe-proto ""))
+    (make-service-info name (vector->list (C-string-vec->Scheme aliases #f))
+		       port protocol)))
 		  
 (define-foreign %service-name->service-info
-  (scheme_serv_name2serv_info 
-   (string name)
-   (string proto))
+  (scheme_serv_name2serv_info (string name) (string proto))
   (to-scheme integer "False_on_zero")
   static-string	 ; service name
   (C char**)     ; alias list
@@ -885,11 +878,11 @@
 (define (protocol-info arg)
   (cond ((string? arg)  (name->protocol-info arg))
 	((integer? arg) (number->protocol-info arg))
-	(else (error "protocol-info: string or integer expected"))))
+	(else (error "protocol-info: string or integer expected ~s" arg))))
 
 (define (number->protocol-info name) 
   (if (not (integer? name))
-      (error "number->protocol-info: integer expected")
+      (error "number->protocol-info: integer expected ~s" name)
       (receive (result name aliases protocol)
 	       (%protocol-port->protocol-info name)
 	 (make-protocol-info name 
@@ -906,7 +899,7 @@
   
 (define (name->protocol-info name)
   (if (not (string? name))
-      (error "name->protocol-info: string expected")
+      (error "name->protocol-info: string expected ~s" name)
       (receive (result name aliases protocol)
 	       (%protocol-name->protocol-info name)
 	 (make-protocol-info name
@@ -955,12 +948,13 @@
 
 ;; also from cig/libcig.scm
 (define-foreign %c-veclen-or-false
-  (veclen ((C long*) c-vec))		; redefining can we open cig-aux?
+  (veclen ((C "const long * ~a") c-vec)); redefining can we open cig-aux?
   desc) ; integer or #f if arg is NULL.
 
 ;; also from cig/libcig.scm
 (define-foreign %set-long-vector-carriers!
-  (set_longvec_carriers (vector-desc svec) ((C char**) cvec))
+  (set_longvec_carriers (vector-desc svec)
+			((C "long const * const * ~a") cvec))
   ignore)
 
 ;; also from cig/libcig.scm
