@@ -28,16 +28,16 @@
 ;;; Common code for FORK/PIPE and %FORK/PIPE.
 (define (really-fork/pipe forker maybe-thunk)
   (receive (r w) (pipe)
-    (let ((pid (forker)))
-      (cond ((zero? pid)
+    (let ((proc (forker)))
+      (cond (proc		; Parent
+	     (close w)
+	     (move->fdes r 0))
+	    (else		; Child
 	     (close r)
 	     (move->fdes w 1)
 	     (if (pair? maybe-thunk)
-		 (call-terminally (car maybe-thunk))))
-	    (else
-	     (close w)
-	     (move->fdes r 0)))
-      pid)))
+		 (call-terminally (car maybe-thunk)))))
+      proc)))
 
 
 ;;; FORK/PIPE with a connection list.
@@ -58,8 +58,16 @@
 		     rev-conns))
 	 (tos (map car rev-conns)))
 
-    (let ((pid (forker)))
-      (cond ((zero? pid)		; Child
+    (let ((proc (forker)))
+      (cond (proc			; Parent
+	     (for-each (lambda (to r/w)
+			 (let ((w (cdr r/w))
+			       (r (car r/w)))
+			   (close w)
+			   (move->fdes r to)))
+		       tos pipes))
+
+	    (else		; Child
 	     (for-each (lambda (from r/w)
 			 (let ((r (car r/w))
 			       (w (cdr r/w)))
@@ -68,16 +76,8 @@
 			   (close w))) ; Unrevealed ports win.
 		       froms pipes)
 	     (if (pair? maybe-thunk)
-		 (call-terminally (car maybe-thunk))))
-
-	    (else			; Parent
-	     (for-each (lambda (to r/w)
-			 (let ((w (cdr r/w))
-			       (r (car r/w)))
-			   (close w)
-			   (move->fdes r to)))
-		       tos pipes)))
-      pid)))
+		 (call-terminally (car maybe-thunk)))))
+      proc)))
 
 (define (tail-pipe a b)
   (fork/pipe a)
@@ -302,7 +302,7 @@
 
 
 (define (temp-file-iterate maker . maybe-template)
-  (let ((template (optional-arg maybe-template (fluid *temp-file-template*))))
+  (let ((template (:optional maybe-template (fluid *temp-file-template*))))
     (let loop ((i 0))
       (if (> i 1000) (error "Can't create temp-file")
 	  (let ((fname (format #f template (number->string i))))
@@ -442,17 +442,17 @@
 ;;;     On eof, return the seeds.
 ;;;     PORT->LIST is just (REDUCE-PORT PORT READ CONS '())
 
-(define (run/port+pid* thunk)
+(define (run/port+proc* thunk)
   (receive (r w) (pipe)
-    (let ((pid (fork (lambda ()
-		       (close r)
-		       (move->fdes w 1)
-		       (with-current-output-port* w thunk)))))
+    (let ((proc (fork (lambda ()
+			(close r)
+			(move->fdes w 1)
+			(with-current-output-port* w thunk)))))
       (close w)
-      (values r pid))))
+      (values r proc))))
 
 (define (run/port* thunk)
-  (receive (port pid) (run/port+pid* thunk)
+  (receive (port proc) (run/port+proc* thunk)
     port))
 
 (define (run/file* thunk)
@@ -472,79 +472,6 @@
 (define (run/strings* thunk)
   (close-after (run/port* thunk) port->string-list))
 
-;;; Pseudo terminals
-
-(define (run/pty* thunk)
-  (receive (pty tty)
-	   (pty-open)
-     (let ((pid (fork (lambda ()
-			(dup->inport  tty 0)
-			(dup->outport tty 1)
-			(dup->outport tty 2)
-			(close tty)
-			(thunk)))))
-       (values pid pty (dup->outport pty)))))
-
-;; returns an open pty and the name of the corresponding tty
-(define (pty-open)
-  (let ((next-pty (make-pty-generator)))
-    (let loop ()
-	(cond ((next-pty) =>
-	       (lambda (pty-name)
-		 ;; what if we have an error opening tty?
-		 ;; supposedly it's agreed to open pty first
-		 ;; but what does the average unix programer know
-		 ;; unix standards. if it works once, it's fine -bri
-		 (with-errno-handler ((errno packet)
-				      ((errno/io errno/acces)
-				       (loop)))
-                   (let ((pty (open-file pty-name open/read+write)))
-		     (values pty
-			     (open-file (pty->tty pty-name) 
-					open/read+write))))))
-	      (else
-	       (error "pty-open: could not open new pty"))))))
-
-;; The following code two pty functions may in fact be system dependant
-;; if so, we'll move it out to the architecture specific directories 
-;; i can't believe this isnt standard or at least a library routine.
-;; oh wait, sysV has libpt.a with routines like this but they are 
-;; undocumented. welcome to the future -bri 
-
-;;; takes a pty string and makes a new string that is the matching tty
-(define (pty->tty pty)
-  (let ((p-pos 5)			; index of p in pty
-	(tty (string-copy pty)))
-    (string-set!  tty p-pos #\t)
-    tty))
-
-;;; a generator for all possible pty names
-(define (make-pty-generator)
-  (let* ((pattern (string-copy"/dev/ptyLN")) ; L=letter N=number
-	 (len (string-length pattern))
-	 (l-pos (- len 2))
-	 (n-pos (- len 1))
-	 ;; from telnetd source in BSD4.4
-	 (letters "pqrstuvwxyzPQRST")
-	 (numbers "0123456789abcdef")
-	 (l-len (string-length letters))
-	 (n-len (string-length numbers))
-	 (l 0)
-	 (n 0))
-    (string-set! pattern l-pos (string-ref letters l)) ; initialize letter
-    (lambda ()
-      (cond ((= n n-len)
-	     (set! l (+ l 1))
-	     (cond ((>= l l-len) #f)
-		   (else
-		    (string-set! pattern l-pos (string-ref letters l))
-		    (string-set! pattern n-pos (string-ref numbers 0)) 
-		    (set! n 1)
-		    pattern)))
-	    (else
-	     (string-set! pattern n-pos (string-ref numbers n)) ; change number
-	     (set! n (+ n 1))
-	     pattern)))))
 
 ;;; Read characters from PORT until EOF, collect into a string.
 
@@ -611,7 +538,7 @@
 		   (lp)))))))
 
 (define (string-filter filter . maybe-buflen)
-  (let* ((buflen (optional-arg maybe-buflen 1024))
+  (let* ((buflen (:optional maybe-buflen 1024))
 	 (buf (make-string buflen)))
     (lambda ()
       (let lp ()
@@ -621,12 +548,45 @@
 				      (substring buf 0 nread)))) ; last one.
 		 (lp))))))))
 
-(define (stdio->stdports thunk)
+(define (y-or-n? question . maybe-eof-value)
+  (let loop ((count *y-or-n-eof-count*))
+    (display question)
+    (display " (y/n)? ")
+    (let ((line (read-line)))
+      (cond ((eof-object? line)
+	     (newline)
+	     (if (= count 0)
+		 (:optional maybe-eof-value (error "EOF in y-or-n?"))
+		 (begin (display "I'll only ask another ")
+			(write count)
+			(display " times.")
+			(newline)
+			(loop (- count 1)))))
+	    ((< (string-length line) 1) (loop count))
+	    ((char=? (string-ref line 0) #\y) #t)
+	    ((char=? (string-ref line 0) #\n) #f)
+	    (else (loop count))))))
+
+(define *y-or-n-eof-count* 100)
+
+
+;;; Stdio/stdport sync procedures
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
+
+(define (stdio->stdports)
+  (set-current-input-port!  (fdes->inport 0))
+  (set-current-output-port! (fdes->outport 1))
+  (set-error-output-port!   (fdes->outport 2)))
+
+(define (with-stdio-ports* thunk)
   (with-current-input-port (fdes->inport 0)
     (with-current-output-port (fdes->outport 1)
       (with-error-output-port (fdes->outport 2)
 	(thunk)))))
 
+(define-simple-syntax (with-stdio-ports body ...)
+  (with-stdio-ports* (lambda () body ...)))
+
 
 (define (stdports->stdio)
   (dup (current-input-port)  0)
@@ -637,25 +597,34 @@
 ;;; Command-line argument access
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
+;;; Some globals.
+(define %command-line '())		; Includes program.
+(define command-line-arguments #f)	; Doesn't include program.
+
+(define (set-command-line-args! args)
+  (set! %command-line args)
+  (set! command-line-arguments (append (cdr args) '())))
+
 (define (arg* arglist n . maybe-default-thunk)
-  (letrec ((oops (lambda () (error "argument out of bounds" arglist n)))
-	   (lp (lambda (al n)
-		 (if (pair? al)
-		     (if (= n 1) (car al)
-			 (lp (cdr al) (- n 1)))
-		     (if (pair? maybe-default-thunk)
-			 ((car maybe-default-thunk))
-			 (oops))))))
+  (let ((oops (lambda () (error "argument out of bounds" arglist n))))
     (if (< n 1) (oops)
-	(lp arglist n))))
+	(let lp ((al arglist) (n n))
+	  (if (pair? al)
+	      (if (= n 1) (car al)
+		  (lp (cdr al) (- n 1)))
+	      (if (and (pair? maybe-default-thunk)
+		       (null? (cdr maybe-default-thunk)))
+		  ((car maybe-default-thunk))
+		  (oops)))))))
 
 (define (arg arglist n . maybe-default)
   (if maybe-default (arg* arglist n (lambda () (car maybe-default)))
       (arg* arglist n)))
 
 (define (argv n . maybe-default)
-  (apply arg %internal-command-line-arguments (+ n 1) maybe-default))
+  (apply arg (cdr %command-line) n maybe-default))
 
+(define (command-line) (append %command-line '()))
 
 ;;; EXEC support
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -709,7 +678,6 @@
 
 	;; Try each directory in PATH-LIST.
 	(let ((argv (list->vector (cons prog (map stringify arglist)))))
-	  (cloexec-unrevealed-ports)
 	  (for-each (lambda (dir)
 		      (let ((binary (string-append dir "/" prog)))
 			(%%exec/errno binary argv env)))
@@ -734,32 +702,48 @@
   (really-fork #f maybe-thunk))
 
 (define (really-fork clear-interactive? maybe-thunk)
-  (let ((pid (%%fork)))
-    (cond ((zero? pid)
-	   (if clear-interactive?
-	       (set-batch-mode?! #t)) ; Children are non-interactive.
-	   (if (pair? maybe-thunk)
-	       (call-terminally (car maybe-thunk)))))
-    pid))
+  ((with-enabled-interrupts 0
+     (let ((pid (%%fork)))
+       (if (zero? pid)				
+
+	   ;; Child
+	   (lambda ()	; Do all this outside the WITH-INTERRUPTS.
+	     (set! reaped-procs '())
+	     (if clear-interactive?
+		 (set-batch-mode?! #t))	; Children are non-interactive.
+	     (and (pair? maybe-thunk)
+		  (call-terminally (car maybe-thunk))))
+
+	   ;; Parent
+	   (let ((proc (new-child-proc pid)))
+	     (lambda () proc)))))))
 
 
 (define (exit . maybe-status)
   (flush-all-ports)
-  (exit/errno (optional-arg  maybe-status 0))
+  (exit/errno (:optional  maybe-status 0))
   (display "The evil undead walk the earth." 2)
   (error "(exit) returned."))
 
-; This definition works for procedures running on top of Unix systems.
+
+;;; The classic T 2.0 primitive.
+;;; This definition works for procedures running on top of Unix systems.
 (define (halts? proc) #t)
 
+
+;;; Low-level init absolutely required for any scsh program.
+
+(define (init-scsh-hindbrain relink-ff?)
+  (if relink-ff? (lookup-all-externals)) ; Re-link C calls.
+  (init-fdports!)
+  (%install-unix-scsh-handlers))
+
+
 ;;; Some globals:
 (define home-directory "")
 (define exec-path-list '())
-(define command-line-arguments #f)
 
-(define (init-scsh quietly?)
-  (lookup-all-externals) ; Re-link C calls.
-  (init-fdports!)
+(define (init-scsh-vars quietly?)
   (set! home-directory
 	(cond ((getenv "HOME") => ensure-file-name-is-nondirectory)
 	      (else (if (not quietly?)
