@@ -16,6 +16,13 @@
   "#include <pwd.h>"
   "#include <unistd.h>"
   ""
+  "/* Make sure foreign-function stubs interface to the C funs correctly: */"
+  "#include \"dirstuff1.h\""
+  "#include \"fdports1.h\""
+  "#include \"select1.h\""
+  "#include \"syscalls1.h\""
+  "#include \"userinfo1.h\""
+  ""
   "extern int errno;"
   ""
   "#define errno_on_zero_or_false(x) ((x) ? SCHFALSE : ENTER_FIXNUM(errno))"
@@ -23,61 +30,49 @@
   "#define False_on_zero(x) ((x) ? ENTER_FIXNUM(x) : SCHFALSE)" ; Not a function.
   "" "")
 
-;;; Macros for converting syscalls that return error codes to ones that
+;;; Macro for converting syscalls that return error codes to ones that
 ;;; raise exceptions on errors.
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 ;;; DEFINE-ERRNO-SYSCALL defines an error-signalling syscall procedure from 
 ;;; one that returns an error code as its first return value -- #f for win,
-;;; errno for lose.
+;;; errno for lose. If the error code is ERRNO/INTR (interrupted syscall),
+;;; we try again.
 ;;;
 ;;; (define-errno-syscall (SYSCALL ARGS) SYSCALL/ERRNO . RET-VALS) ==>
 ;;;
 ;;; (define (SYSCALL . ARGS)
 ;;;   (receive (err . RET-VALS) (SYSCALL/ERRNO . ARGS)
-;;;     (if err (errno-error err SYSCALL)
-;;;     	(values . RET-VALS))))
-    
+;;;     (cond ((not err) (values . RET-VALS))		; Win
+;;;           ((= err errno/intr) (SYSCALL . ARGS))	; Retry
+;;;           (else (errno-error err SYSCALL . ARGS))))); Lose
+
 (define-syntax define-errno-syscall
   (syntax-rules ()
     ((define-errno-syscall (syscall arg ...) syscall/errno
        ret-val ...)
      (define (syscall arg ...)
        (receive (err ret-val ...) (syscall/errno arg ...)
-         (if err (errno-error err syscall)
-             (values ret-val ...)))))
+         (cond ((not err) (values ret-val ...))			; Win
+	       ((= err errno/intr) (syscall arg ...))		; Retry
+	       (else (errno-error err syscall arg ...))))))	; Lose
 
     ;;; This case handles rest args
     ((define-errno-syscall (syscall . args) syscall/errno
        ret-val ...)
      (define (syscall . args)
-       (receive (err ret-val ...) (apply syscall/errno . args)
-         (if err (errno-error err syscall)
-             (values ret-val ...)))))))
+       (receive (err ret-val ...) (apply syscall/errno args)
+         (cond ((not err) (values ret-val ...))			; Win
+	       ((= err errno/intr) (apply syscall args))	; Retry
+	       (else (apply errno-error err syscall args)))))))); Lose
 
-;;; DEFINE-SIMPLE-ERRNO-SYSCALL is for the simple case of a system call
-;;; that returns no interesting value other than its errno code (or #f
-;;; for success). This is most syscalls.
-;;;
-;;; (define-simple-errno-syscall (SYSCALL . ARGS) SYSCALL/ERRNO) =>
-;;;
-;;; (define (SYSCALL . ARGS)
-;;;   (cond ((SYSCALL/ERRNO . ARGS) => 
-;;;          (lambda (err) (errno-error err SYSCALL)))))
-
-(define-syntax define-simple-errno-syscall
-  (syntax-rules ()
-    ((define-simple-errno-syscall (syscall arg ...) syscall/errno)
-     (define (syscall arg ...)
-       (cond ((syscall/errno arg ...) => 
-              (lambda (err) (errno-error err syscall))))))
-
-
-    ;; This case handles rest args
-    ((define-simple-errno-syscall (syscall . args) syscall/errno)
-     (define (syscall . args)
-       (cond ((apply syscall/errno . args) => 
-              (lambda (err) (errno-error err syscall))))))))
+;;; By the way, it would be better to insert a (LET LP () ...) for the
+;;; the errno/intr retries, instead of calling the top-level definition
+;;; (because in Scheme you have to allow for the fact that top-level
+;;; defns can be re-defined, so the compiler can't just turn it into a
+;;; jump), but the brain-dead S48 byte-compiler will cons a closure for
+;;; the LP loop, which means that little syscalls like read-char can cons
+;;; like crazy. So I'm doing it this way. Ech.
 
 
 ;;; Process
@@ -90,14 +85,12 @@
   integer)
 
 (define (%%exec prog argv env)
-  (errno-error (%%exec/errno prog argv env) %exec)) ; cute.
+  (errno-error (%%exec/errno prog argv env) %exec prog argv env)) ; cute.
 
 (define (%exec prog arg-list env)
   (let ((argv (mapv! stringify (list->vector arg-list)))
 	(prog (stringify prog))
-	(env (if (eq? env #t) #t
-		 (list->vector (alist->env-list env)))))
-    (cloexec-unrevealed-ports)
+	(env (if (eq? env #t) #t (alist->env-vec env))))
     (%%exec prog argv env)))
 
 
@@ -110,7 +103,7 @@
   ignore)
 
 (define (%exit . maybe-status)
-  (%exit/errno (optional-arg maybe-status 0))
+  (%exit/errno (:optional maybe-status 0))
   (error "Yikes! %exit returned."))
 
 
@@ -118,37 +111,29 @@
   (multi-rep (to-scheme pid_t errno_or_false)
              pid_t))
 
-(define-errno-syscall (%%fork) %%fork/errno
+;;; If the fork fails, and we are doing early zombie reaping, then reap
+;;; some zombies to try and free up a some space in the process table,
+;;; and try again.
+;;;
+;;; This ugly little hack will have to stay in until I do early
+;;; zombie reaping with SIGCHLD interrupts.
+
+(define (%%fork-with-retry/errno)
+  (receive (err pid) (%%fork/errno)
+    (cond ((and err (eq? 'early (autoreap-policy)))
+	   (reap-zombies)
+	   (%%fork/errno))
+	  (else (values err pid)))))
+
+(define-errno-syscall (%%fork) %%fork-with-retry/errno
   pid)
 
-
-;;; Tricky direct interface to wait(2).
-
-(define-foreign %wait-any/errno (wait)
-  (multi-rep (to-scheme integer errno_or_false) ; error flag
-	     integer) ; pid
-  integer) ; pid's status.
-
-;;; [#f #f] on no reportable child, otw [pid, status].
-(define (wait-any)
-  (receive (err pid status) (%wait-any/errno)
-    (if err
-	(if (= err errno/child)
-	    (values #f #f)
-	    (errno-error err wait-any))
-	(values pid status))))
-
+;;; Posix waitpid(2) call.
 (define-foreign %wait-pid/errno (wait_pid (integer pid) (integer options))
-  (to-scheme integer errno_or_false) ; error flag
+  desc		; errno or #f
+  integer  ; process' id
   integer) ; process' status
 
-;;; Maybe change errno-syscall to be a CALL, not a PROCEDURE...
-(define (%wait-pid pid . maybe-options)
-  (let ((options (optional-arg maybe-options 0)))
-    (receive (err status) (%wait-pid/errno pid options)
-      (if err (errno-error err %wait-pid)
-	  status))))
-
 
 ;;; Miscellaneous process state
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -159,10 +144,10 @@
   (chdir (string directory))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (%chdir dir) %chdir/errno)
+(define-errno-syscall (%chdir dir) %chdir/errno)
 
 (define (chdir . maybe-dir)
-  (let ((dir (optional-arg maybe-dir ".")))
+  (let ((dir (:optional maybe-dir (home-dir))))
     (%chdir (ensure-file-name-is-nondirectory dir))))
 
 
@@ -179,10 +164,10 @@
 (define-foreign user-gid  (getgid) gid_t)
 (define-foreign user-effective-gid (getegid) gid_t)
 
-(define-foreign set-gid/errno (setgid (gid_t id))
+(define-foreign set-gid/errno (setgid (gid_t id)) no-declare ; for SunOS 4.x 
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (set-gid gid) set-gid/errno)
+(define-errno-syscall (set-gid gid) set-gid/errno)
 
 (define-foreign %num-supplementary-gids/errno (num_supp_groups)
   (multi-rep (to-scheme integer errno_or_false)
@@ -206,10 +191,10 @@
 (define-foreign user-uid		(getuid)  uid_t)
 (define-foreign user-effective-uid 	(geteuid) uid_t)
 
-(define-foreign set-uid/errno (setuid (uid_t id))
+(define-foreign set-uid/errno (setuid (uid_t id)) no-declare ; for SunOS 4.x
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (set-uid uid_t) set-uid/errno)
+(define-errno-syscall (set-uid uid_t) set-uid/errno)
 
 (define-foreign %user-login-name (my_username)
   static-string)
@@ -224,22 +209,33 @@
 (define-foreign parent-pid (getppid) pid_t)
 
 
-;;; Process Groups   -dalbertz
+;;; Process groups and session ids
+;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define-foreign process-group (getpgrp) pid_t)
 (define-foreign %set-process-group/errno
   (setpgid (pid_t pid) (pid_t groupid))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (%set-process-group pid pgrp)
+(define-errno-syscall (%set-process-group pid pgrp)
   %set-process-group/errno)
 
+
 (define (set-process-group arg1 . maybe-arg2)
   (receive (pid pgrp) (if (null? maybe-arg2)
 			  (values (pid) arg1)
 			  (values arg1 (car maybe-arg2)))
 	   (%set-process-group pid pgrp)))
 
+
+(define-foreign become-session-leader/errno (setsid)
+  (multi-rep (to-scheme pid_t errno_or_false)
+	     pid_t))
+
+(define-errno-syscall (become-session-leader) become-session-leader/errno
+  sid)
+
+
 ;;; UMASK
 
 (define-foreign set-umask (umask (mode_t mask)) no-declare ; integer on SunOS
@@ -254,6 +250,7 @@
 ;;; PROCESS TIMES
 
 ;;; OOPS: The POSIX times() has a mildly useful ret value we are throwing away.
+;;; OOPS: The ret values should be clock_t, not int, but cig can't handle it.
 
 (define-foreign process-times/errno (process_times)
   (to-scheme integer errno_or_false)
@@ -265,6 +262,8 @@
 (define-errno-syscall (process-times) process-times/errno
   utime stime cutime cstime)
 
+(define-foreign cpu-ticks/sec (cpu_clock_ticks_per_sec) integer)
+
 ;;; File system
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
@@ -289,7 +288,7 @@
   no-declare ; Workaround for AIX bug.
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (set-file-mode thing mode)
+(define-errno-syscall (set-file-mode thing mode)
   (lambda (thing mode)
     (generic-file-op thing
 		     (lambda (fd)    (set-fdes-mode/errno fd    mode))
@@ -305,17 +304,17 @@
   (fchown (integer fd) (uid_t uid) (gid_t gid))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (set-file-owner thing uid)
+(define-errno-syscall (set-file-owner thing uid)
   (lambda (thing uid)
     (generic-file-op thing
 		     (lambda (fd)    (set-fdes-uid&gid/errno fd    uid -1))
 		     (lambda (fname) (set-file-uid&gid/errno fname uid -1)))))
 
-(define-simple-errno-syscall (set-file-group thing gid)
+(define-errno-syscall (set-file-group thing gid)
   (lambda (thing gid)
     (generic-file-op thing
-		     (lambda (fd)    (set-fdes-uid&gid/errno fd    gid -1))
-		     (lambda (fname) (set-file-uid&gid/errno fname gid -1)))))
+		     (lambda (fd)    (set-fdes-uid&gid/errno fd    -1 gid))
+		     (lambda (fname) (set-file-uid&gid/errno fname -1 gid)))))
 
 
 ;;; Uses real uid and gid, not effective. I don't use this anywhere.
@@ -342,7 +341,7 @@
   (link (string original-name) (string new-name))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (create-hard-link original-name new-name)
+(define-errno-syscall (create-hard-link original-name new-name)
   create-hard-link/errno)
 
 
@@ -350,7 +349,7 @@
   no-declare ; integer on SunOS
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (create-fifo path mode) create-fifo/errno)
+(define-errno-syscall (create-fifo path mode) create-fifo/errno)
 
 
 (define-foreign create-directory/errno
@@ -358,10 +357,11 @@
   (to-scheme integer errno_or_false))
 
 (define (create-directory path . maybe-mode)
-  (let ((mode (optional-arg maybe-mode #o777))
+  (let ((mode (:optional maybe-mode #o777))
 	(fname (ensure-file-name-is-nondirectory path)))
     (cond ((create-directory/errno fname mode) =>
-           (lambda (err) (if err (errno-error err create-directory)))))))
+           (lambda (err)
+	     (if err (errno-error err create-directory path mode)))))))
 
 
 (define-foreign read-symlink/errno (scm_readlink (string path))
@@ -376,7 +376,7 @@
   (rename (string old-name) (string new-name))
   (to-scheme integer errno_or_false))
   
-(define-simple-errno-syscall (%rename-file old-name new-name)
+(define-errno-syscall (%rename-file old-name new-name)
   %rename-file/errno)
 
 
@@ -384,7 +384,32 @@
   (rmdir (string path))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (delete-directory path) delete-directory/errno)
+(define-errno-syscall (delete-directory path) delete-directory/errno)
+
+
+(define-foreign %utime/errno (scm_utime (string path)
+					(integer ac_hi) (integer ac_lo)
+					(integer m_hi)  (integer m_lo))
+  (to-scheme integer errno_or_false))
+
+(define-foreign %utime-now/errno (scm_utime_now (string path))
+  (to-scheme integer errno_or_false))
+					
+
+;;; (SET-FILE-TIMES/ERRNO path [access-time mod-time])
+
+(define (set-file-times/errno path . maybe-times)
+  (if (pair? maybe-times)
+      (let* ((access-time (real->exact-integer (car maybe-times)))
+	     (mod-time (if (pair? (cddr maybe-times))
+			   (error "Too many arguments to set-file-times/errno"
+				  (cons path maybe-times))
+			   (real->exact-integer (cadr maybe-times)))))
+	(%utime/errno path (hi8 access-time) (lo24 access-time)
+		           (hi8 mod-time)    (lo24 mod-time)))
+      (%utime-now/errno path)))
+
+(define-errno-syscall (set-file-times . args) set-file-times/errno)
 
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -418,7 +443,7 @@
 
 
 ;;; Should be redone to return multiple-values.
-(define (file-attributes/errno fd/port/fname chase?)
+(define (file-info/errno fd/port/fname chase?)
   (let ((ans-vec (make-vector 14))
 	(time-hack (lambda (lo-24 hi-8)
 		     (let ((val (+ (arithmetic-shift hi-8 24) lo-24)))
@@ -454,30 +479,40 @@
 					   (time-hack (vector-ref ans-vec 12)
 						      (vector-ref ans-vec 13))))))))
 		
-(define (file-attributes fd/port/fname . maybe-chase?)
-  (receive (err info) (file-attributes/errno fd/port/fname 
-					     (optional-arg maybe-chase? #t))
-    (if err (errno-error err file-attributes)
-	info)))
+(define (file-info fd/port/fname . maybe-chase?)
+  (let ((chase? (:optional maybe-chase? #t)))
+    (receive (err info) (file-info/errno fd/port/fname chase?)
+      (if err (errno-error err file-info fd/port/fname chase?)
+	  info))))
 
 
+(define file-attributes
+  (deprecated-proc file-info "file-attributes" "Use file-info instead."))
+
+
+;;; "no-declare" as there is no agreement among the OS's as to whether or not
+;;; the OLD-NAME arg is "const". It *should* be const.
+
 (define-foreign create-symlink/errno
-  (symlink (string old-name) (string new-name))
+  (symlink (string old-name) (string new-name))	no-declare
   (to-scheme integer errno_or_false))
   
-;(define-simple-errno-syscall (create-symlink old-name new-name)
+;(define-errno-syscall (create-symlink old-name new-name)
 ;  create-symlink/errno)
 
 
+;;; "no-declare" as there is no agreement among the OS's as to whether or not
+;;; the PATH arg is "const". It *should* be const.
+
 (define-foreign truncate-file/errno
-  (truncate (string path) (off_t length)) 
+  (truncate (string path) (off_t length)) 	no-declare
   (to-scheme integer errno_or_false))
 
 (define-foreign truncate-fdes/errno
   (ftruncate (integer fd) (off_t length))   no-declare ; Indigo bogosity.
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (truncate-file path length)
+(define-errno-syscall (truncate-file path length)
   (lambda (thing length)
     (generic-file-op thing
 		     (lambda (fd)    (truncate-fdes/errno fd    length))
@@ -488,20 +523,21 @@
   (unlink (string path))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (delete-file path) delete-file/errno)
+(define-errno-syscall (delete-file path) delete-file/errno)
 
 
 (define-foreign sync-file/errno (fsync (integer fd))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (sync-file fd/port)
+(define-errno-syscall (sync-file fd/port)
   (lambda (fd/port)
     (if (output-port? fd/port) (force-output fd/port))
-    (call/fdes fd/port sync-file/errno)))
+    (sleazy-call/fdes fd/port sync-file/errno)))
 
 
 ;;; Amazingly bogus syscall -- doesn't *actually* sync the filesys.
-(define-foreign sync-file-system (sync) ignore)
+(define-foreign sync-file-system (sync) no-declare ; Linux sux - says int
+  ignore)
 
 
 ;;; I/O
@@ -510,8 +546,14 @@
 (define-foreign %close-fdes/errno (close (integer fd))
   (to-scheme integer "errno_or_false"))
 
-(define-simple-errno-syscall (%close-fdes fd) %close-fdes/errno)
-
+(define (%close-fdes fd)
+  (let lp ()
+    (let ((errno (%close-fdes/errno fd)))
+      (cond ((not errno) 	  #t)	; Successful close.
+	    ((= errno errno/badf) #f)	; File descriptor already closed.
+	    ((= errno errno/intr) (lp))	; Retry.
+	    (else
+	     (errno-error errno %close-fdes fd))))))	; You lose.
 
 (define-foreign %dup/errno
   (dup (integer fd))
@@ -530,25 +572,32 @@
    new-fd)
 
 
-(define-foreign %file-seek/errno
+(define-foreign %fd-seek/errno
   (lseek (integer fd) (off_t offset) (integer whence))
   (multi-rep (to-scheme off_t errno_or_false)
 	     off_t))
 
-(define-errno-syscall (%file-seek fd offset whence) %file-seek/errno
-   position)
+
 
 (define seek/set 0)			;Unix codes for "whence"
 (define seek/delta 1)
 (define seek/end 2)
 
-(define (file-seek fd/port offset whence)
-  (if (output-port? fd/port) (force-output fd/port))
-  (call/fdes fd/port (lambda (fd) 
-		       (receive (err info)
-				(%file-seek/errno fd offset whence)
-				(if err (errno-error err file-seek)
-				    info)))))
+(define (seek fd/port offset . maybe-whence)
+  (let ((whence (:optional maybe-whence seek/set)))
+    (receive (err cursor)  
+	((if (integer? fd/port) %fd-seek/errno %fdport-seek/errno)
+	 fd/port
+	 offset
+	 whence)
+      (if err (errno-error err seek fd/port offset whence) cursor))))
+
+(define (tell fd/port)
+  (receive (err offset) (if (integer? fd/port)
+			    (%fd-seek/errno fd/port 0 seek/delta) ; seek(fd)
+			    (%fdport-tell/errno fd/port))	  ; ftell(f)
+    (if err (errno-error err tell fd/port) offset)))
+
 
 (define-foreign %char-ready-fdes?/errno
   (char_ready_fdes (integer fd))
@@ -572,7 +621,7 @@
    fd)
 
 (define (open-fdes path flags . maybe-mode) ; mode defaults to 0666
-  (%open path flags (optional-arg maybe-mode #o666)))
+  (%open path flags (:optional maybe-mode #o666)))
 
 
 (define-foreign pipe-fdes/errno (scheme_pipe)
@@ -597,13 +646,13 @@
 
 (define (read-fdes-char fd)
   (let ((c (%read-fdes-char fd)))
-    (if (integer? c) (errno-error c read-fdes-char) c)))
+    (if (integer? c) (errno-error c read-fdes-char fd) c)))
 
 
 (define-foreign write-fdes-char/errno (write_fdes_char (char char) (integer fd))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (write-fdes-char char fd) write-fdes-char/errno)
+(define-errno-syscall (write-fdes-char char fd) write-fdes-char/errno)
 
 
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -626,26 +675,35 @@
 	     integer))
 
 
-
 ;;; Signals (rather incomplete)
 ;;; ---------------------------
 
-(define-foreign signal-process/errno
+(define-foreign signal-pid/errno
   (kill (pid_t pid) (integer signal))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (signal-process pid signal) signal-process/errno)
+(define-errno-syscall (signal-pid pid signal) signal-pid/errno)
 
-(define (signal-procgroup proc-group signal)
-  (signal-process (- proc-group) signal))
+(define (signal-process proc signal)
+  (signal-pid (cond ((proc? proc)    (proc:pid proc))
+		    ((integer? proc) proc)
+		    (else (error "Illegal proc passed to signal-process" proc)))
+	      signal))
+
+(define (signal-process-group proc-group signal)
+  (signal-pid (- (cond ((proc? proc-group)    (proc:pid proc-group))
+		       ((integer? proc-group) proc-group)
+		       (else (error "Illegal proc passed to signal-process-group"
+                                    proc-group))))
+              signal))
 
 ;;; SunOS, not POSIX:
-;;; (define-foreign signal-procgroup/errno
+;;; (define-foreign signal-process-group/errno
 ;;;   (killpg (integer proc-group) (integer signal))
 ;;;   (to-scheme integer errno_or_false))
 ;;; 
-;;; (define-simple-errno-syscall (signal-procgroup proc-group signal)
-;;;   signal-procgroup/errno)
+;;; (define-errno-syscall (signal-process-group proc-group signal)
+;;;   signal-process-group/errno)
 
 (define-foreign pause-until-interrupt (pause) no-declare ignore)
 
@@ -655,7 +713,11 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define-record user-info
-  name uid gid home-dir shell)
+  name uid gid home-dir shell
+
+  ;; Make user-info records print like #{user-info shivers}.
+  ((disclose ui)
+   (list "user-info" (user-info:name ui))))
 
 (define-foreign %uid->user-info (user_info_uid (uid_t uid))
   bool		; win?
@@ -705,7 +767,10 @@
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
 (define-record group-info
-  name gid members)
+  name gid members
+
+  ;; Make group-info records print like #{group-info wheel}.
+  ((disclose gi) (list "group-info" (group-info:name gi))))
 
 ;;; These guys return static structs, so they aren't reentrant.
 ;;; Must be fixed for threaded version.
@@ -766,17 +831,17 @@
 ;;; Sorts them in place by the Unix filename order: ., .., dotfiles, others.
 
 (define-foreign %sort-file-vector
-  (scm_sort_filevec ((C char**) cvec)
+  (scm_sort_filevec ((C "const char** ~a") cvec)
 		    (integer    veclen))
   ignore)
 
 (define (directory-files . args)
-  (receive (dir dotfiles?)
-           (parse-optionals args "." #f)
+  (let-optionals args ((dir       ".")
+		       (dotfiles? #f))
     (check-arg string? dir directory-files)
     (receive (err cvec numfiles)
 	     (%open-dir (ensure-file-name-is-nondirectory dir))
-      (if err (errno-error err directory-files))
+      (if err (errno-error err directory-files dir dotfiles?))
       (%sort-file-vector cvec numfiles)
       (let ((files (vector->list (C-string-vec->Scheme&free cvec numfiles))))
 	(if dotfiles? files
@@ -784,20 +849,22 @@
 		    files))))))
 
 ;;; I do this one in C, I'm not sure why:
+;;; It is used by MATCH-FILES.
 
 (define-foreign %filter-C-strings!
-  (filter_stringvec (string regexp) ((C char**) cvec))
-  static-string	; error message -- ""  if no error.
+  (filter_stringvec (string pattern) ((C "char const ** ~a") cvec))
+  static-string	; error message -- #f if no error.
   integer)	; number of files that pass the filter.
 
+
 (define (match-files regexp . maybe-dir)
-  (let ((dir (optional-arg maybe-dir ".")))
+  (let ((dir (:optional maybe-dir ".")))
     (check-arg string? dir match-files)
     (receive (err cvec numfiles)
 	     (%open-dir (ensure-file-name-is-nondirectory dir))
-      (if err (errno-error err match-files))
+      (if err (errno-error err match-files regexp dir))
       (receive (err numfiles) (%filter-C-strings! regexp cvec)
-	(if (not (equal? err "")) (error err match-files))
+	(if err (error err match-files))
 	(%sort-file-vector cvec numfiles)
 	(let ((files (C-string-vec->Scheme&free cvec numfiles)))
 	  (vector->list files))))))
@@ -820,9 +887,10 @@
 			   cons))
        env-list))
 
-(define (alist->env-list alist)
-  (map (lambda (var.val) (string-append (car var.val) "=" (cdr var.val)))
-       alist))
+(define (alist->env-vec alist)
+  (list->vector (map (lambda (var.val)
+		       (string-append (car var.val) "=" (cdr var.val)))
+		     alist)))
 
 ;;; ENV->ALIST
 
@@ -842,10 +910,10 @@
   (install_env (vector-desc env-vec))
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (%install-env env-vec) %install-env/errno)
+(define-errno-syscall (%install-env env-vec) %install-env/errno)
 
 (define (alist->env alist)
-  (%install-env (alist->env-list alist)))
+  (%install-env (alist->env-vec alist)))
 
 ;;; GETENV, PUTENV, SETENV
 
@@ -885,7 +953,15 @@
 (define-foreign close-fdport*/errno (close_fdport (desc data))
   (to-scheme integer "False_on_zero")) 	; Win: #f, lose: errno
 
-(define-simple-errno-syscall (close-fdport* data) close-fdport*/errno)
+(define (close-fdport* data) 
+  (let lp ()
+    (let ((errno (close-fdport*/errno data)))
+      (cond ((not errno) 	  #t)	; Successful close.
+	    ((= errno errno/badf) #f)	; File descriptor already closed.
+	    ((= errno errno/intr) (lp))	; Retry.
+	    (else
+	     (errno-error errno close-fdport* data))))))    ; You lose.
+
 
 
 (define-foreign %fdport*-read-char/errno (fdport_getchar (desc data))
@@ -893,8 +969,11 @@
 
 (define (%fdport*-read-char data)
   (let ((c (%fdport*-read-char/errno data)))
-    (if (integer? c) (errno-error c %fdport*-read-char)
-	(or c eof-object))))
+    (if (integer? c)
+	(if (= c errno/intr)
+	    (%fdport*-read-char data)			; Retry
+	    (errno-error c %fdport*-read-char data))	; Lose
+	(or c eof-object))))				; Win
 
 
 (define-foreign %fdport*-char-ready?/errno
@@ -903,38 +982,61 @@
 
 (define (%fdport*-char-ready? data)
   (let ((val (%fdport*-char-ready?/errno data)))
-    (if (integer? val) (errno-error val %fdport*-char-ready?)
+    (if (integer? val) (errno-error val %fdport*-char-ready? data)
 	val)))
 
-
 (define-foreign %fdport*-write-char/errno
   (fdport_putchar (desc data) (char c))
   (to-scheme integer "False_on_zero")) 	; Win: #f, lose: errno
 
-(define-simple-errno-syscall (%fdport*-write-char desc c)
+(define-errno-syscall (%fdport*-write-char desc c)
   %fdport*-write-char/errno)
 
-
 (define-foreign flush-fdport*/errno (flush_fdport (desc data))
   (to-scheme integer "False_on_zero")) 	; Win: #f, lose: errno
 
-(define-simple-errno-syscall (flush-fdport* data) flush-fdport*/errno)
+;;; FLUSH-FDPORT* isn't defined with DEFINE-ERRNO-SYSCALL because that would
+;;; return 0 values, which blows up S48's extended-port machinery. This
+;;; version returns #f.
+;;; ???
+
+(define (flush-fdport* data)
+  (cond ((flush-fdport*/errno data) =>
+         (lambda (err) (if (= err errno/intr)
+			   (flush-fdport* data)
+			   (errno-error err flush-fdport* data))))
+	(else #f)))
 
 (define-foreign flush-all-ports/errno (flush_all_ports)
   (to-scheme integer errno_or_false))
 
-(define-simple-errno-syscall (flush-all-ports)
+(define-errno-syscall (flush-all-ports)
   flush-all-ports/errno)
 
+(define-foreign %fdport*-seek/errno
+  (seek_fdport (desc data) (off_t offset) (integer whence))
+  (to-scheme integer "False_on_zero")	; errno
+  integer)				; new position
+
+(define-foreign %fdport*-tell/errno
+  (tell_fdport (desc data))
+  (to-scheme integer "False_on_zero")	; errno
+  integer)
+
+(define-foreign %fdport*-set-buffering/errno
+  (set_fdbuf (desc data) (integer policy) (integer size))
+  (to-scheme integer "False_on_zero"))	; errno
+
+(define-foreign %set-cloexec (set_cloexec (integer fd) (bool val))
+  (to-scheme integer "errno_or_false"))
+
 (define-foreign %init-fdports! (init_fdports) ignore)
 
-(define-foreign cloexec-unrevealed-ports (cloexec_unrevealed) ignore)
-
 (define-foreign %install-port/errno
-  (install_port (integer fd) (desc port))
+  (install_port (integer fd) (desc port) (integer revealed))
   (to-scheme integer "False_on_zero")) ; Win: #f, lose: errno
   
-(define-simple-errno-syscall (%install-port fd port) %install-port/errno)
+(define-errno-syscall (%install-port fd port revealed) %install-port/errno)
 
 
 (define-foreign %maybe-fdes->port (maybe_fdes2port (integer fd))
@@ -977,12 +1079,33 @@
   (to-scheme integer errno_or_false))
 
 (define-errno-syscall (%fcntl-read fd command) %fcntl-read/errno value)
-(define-simple-errno-syscall (%fcntl-write fd command val) %fcntl-write/errno)
+(define-errno-syscall (%fcntl-write fd command val) %fcntl-write/errno)
 
-(define (i/o-flags fd/port)
-  (call/fdes fd/port
-    (lambda (fd) (%fcntl-read fd fcntl/get-file-flags))))
+;;; fcntl()'s F_GETFD and F_SETFD.  Note that the SLEAZY- prefix on the
+;;; CALL/FDES isn't an optimisation; it's *required* for the correct behaviour
+;;; of these procedures. Straight CALL/FDES modifies unrevealed file
+;;; descriptors by clearing their CLOEXEC bit when it reveals them -- so it
+;;; would interfere with the reading and writing of that bit!
 
+(define (fdes-flags fd/port)
+  (sleazy-call/fdes fd/port
+    (lambda (fd) (%fcntl-read fd fcntl/get-fdes-flags))))
+
+(define (set-fdes-flags fd/port flags)
+  (sleazy-call/fdes fd/port
+    (lambda (fd) (%fcntl-write fd fcntl/set-fdes-flags flags))))
+
+;;; fcntl()'s F_GETFL and F_SETFL.
+;;; Get: Returns open flags + get-status flags (below)
+;;; Set: append, sync, async, nbio, nonblocking, no-delay
+
+(define (fdes-status fd/port)
+  (sleazy-call/fdes fd/port
+    (lambda (fd) (%fcntl-read fd fcntl/get-status-flags))))
+
+(define (set-fdes-status fd/port flags)
+  (sleazy-call/fdes fd/port
+    (lambda (fd) (%fcntl-write fd fcntl/set-status-flags flags))))
 
 ;;; Miscellaneous
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
@@ -993,7 +1116,18 @@
 ; De-released -- not POSIX and not on SGI systems.
 ; (define-foreign usleep (usleep (integer usecs)) integer)
 
-(define-foreign sleep (sleep (uint_t secs)) uint_t)
+(define (sleep secs) (sleep-until (+ secs (time))))
+
+(define (sleep-until when)
+  (let* ((when (floor when))	; Painful to do real->int in Scheme.
+	 (when (if (exact? when) when (inexact->exact when)))
+	 (h (hi8 when)) (l (lo24 when)))
+    (let lp ()
+      (or (%sleep-until h l) (lp)))))
+
+(define-foreign %sleep-until (sleep_until (fixnum hi)
+					  (fixnum lo))
+  desc)
 
 (define-foreign %gethostname (scm_gethostname)
   static-string)
