@@ -1,6 +1,7 @@
 extern scheme_value scheme_time(int *hi_secs, int *lo_secs);
 
-extern scheme_value scheme_utime(int *hi_secs, int *lo_secs, int *usecs);
+extern scheme_value time_plus_ticks(int *hi_secs,  int *lo_secs,
+				    int *hi_ticks, int *lo_ticks);
 
 extern scheme_value time2date(int hi_secs, int lo_secs, scheme_value zone,
 			      int *sec, int *min, int *hour,
