@@ -60,11 +60,6 @@
 
 #include "sysdep.h"
 
-#if defined(HAVE_HARRIS)
-#undef HAVE_NLIST
-#undef USCORE
-#endif
-
 #include <stdio.h>
 #include <stdlib.h>		/* for getenv(), etc. (POSIX?/ANSI) */
 #include <string.h>		/* for strncpy(), etc. (POSIX/ANSI) */
@@ -95,7 +90,7 @@
 #  include <sys/timeb.h>	/* for ftime() */
 #endif
 
-#if defined(HAVE_NLIST)
+#if defined(HAVE_NLIST_H)
 #  include <nlist.h>		/* conforms to "SVID2", whatever that is */
 #endif
 
@@ -108,9 +103,21 @@ extern long Spending_interruptsS;
 
 /* Signal handlers */
 
+/* Note that we turn off SIGPIPE interrupts in a funny way -- we don't
+** set the handler to SIGIGN, we set it to a no-op function. This is so
+** that when we exec() some other program, the handler will be reset to
+** its default, rather than to SIGIGN.
+**
+** We don't use SIGPIPE, because all write() calls in scsh correctly check
+** for error returns, and raise an error exception. Sleazy Unix C programs,
+** however, need to get shot down when writing to a closed pipe or socket.
+**     -Olin 4/97
+*/
+
+
 static RETSIGTYPE
 when_keyboard_interrupt(sig, code, scp)
-     int sig, code; 
+     int sig, code;
      struct sigcontext *scp;
 {
   Spending_interruptsS |= (1 << INTERRUPT_KEYBOARD);
@@ -122,13 +129,21 @@ when_keyboard_interrupt(sig, code, scp)
 
 static RETSIGTYPE
 when_alarm_interrupt(sig, code, scp)
-     int sig, code; 
+     int sig, code;
      struct sigcontext *scp;
 {
   Spending_interruptsS |= (1 << INTERRUPT_ALARM);
   return;
 }
 
+static RETSIGTYPE
+when_pipe_interrupt(sig, code, scp)
+     int sig, code;
+     struct sigcontext *scp;
+{
+  return;
+}
+
 
 /* OS-dependent initialization */
 
@@ -147,6 +162,14 @@ sysdep_init()
   sigemptyset(&alarm_action.sa_mask);
 
   sigaction(SIGINT, &keyboard_action, NULL);
+
+  /* SIGPIPE's are bogus. -Olin */
+  {struct sigaction sa;
+   sa.sa_handler = when_pipe_interrupt;
+   sigemptyset(&sa.sa_mask);
+   sa.sa_flags = 0;
+   sigaction(SIGPIPE, &sa, NULL);
+   }
 }
 
 /* ---------------------------------------- */
@@ -155,7 +178,6 @@ sysdep_init()
 int
 char_ready_p( FILE* stream )
 {
-  fd_set readfds;
   struct timeval timeout;
   static int warnedp = 0;
 
@@ -173,7 +195,7 @@ char_ready_p( FILE* stream )
 #elif defined(__linux__)
   if (stream->_IO_read_ptr < stream->_IO_read_end)
     return 1;
-#elif defined(__NetBSD__)
+#elif defined(BSD) && (BSD >= 199306)
   if (stream->_r)
     return 1;
 #else
@@ -192,13 +214,12 @@ char_ready_p( FILE* stream )
 
   /* Nothing in the buffer.  Find out whether a read would block. */
 #if defined(HAVE_SELECT)
-  FD_ZERO(&readfds);
-  FD_SET(fileno(stream), &readfds);
-  timerclear(&timeout);
-  return select(FD_SETSIZE,
-		(fd_set_param *)
-		&readfds,
-		NULL, NULL, &timeout);
+  {  fd_set readfds;
+     FD_ZERO(&readfds);
+     FD_SET(fileno(stream), &readfds);
+     timerclear(&timeout);
+     return select1(FD_SETSIZE, &readfds, NULL, NULL, &timeout);
+     }
 #else /* No select() - but there will generally be some other way to do this.*/
   if (!warnedp) {
     fprintf(stderr, "Warning: incomplete char-ready? implementation.\n");
@@ -478,7 +499,7 @@ fill_the_table(void)
     fprintf(stderr, "Error during external name lookup\n");
     return;
   }
-  
+
   if ((f = ldopen(reloc_info_file, NULL)) == NULL) {
     fprintf(stderr, "Error during external name lookup\n");
     return;
@@ -521,7 +542,7 @@ fill_the_table(void)
 
   if (cur_entry)
     cur_entry->next = NULL;
-	
+
   ldclose(f);
 }
 
@@ -537,7 +558,7 @@ really_lookup_external_name(char *name)
   }
   return(0);
 }
-      
+
 long
 lookup_external_name(char *name, long *location)
 {
@@ -567,6 +588,19 @@ lookup_external_name(char *name, long *location)
 
 #else /* i.e. #if !defined(_AIX) */
 
+#if defined(__hpux)
+#include <dl.h>
+lookup_external_name (char *name, long *location )
+{
+    int ret;
+    shl_t shlh = PROG_HANDLE;
+    *location = 0xdeadbeef;
+    ret = shl_findsym (&shlh, name, TYPE_UNDEFINED, location);
+    return (!ret);
+}
+
+#else /* i.e. #if !defined(__hpux) && !defined(_AIX) */
+
 #if !defined(NLIST_HAS_N_NAME)
 #define n_name	n_un.n_name
 #endif
@@ -618,7 +652,7 @@ lookup_external_name(char *name, long *location)
 
   name_list[0].n_name = nm;
   name_list[1].n_name = 0;
-  
+
   status = nlist(reloc_info_file, name_list);
 
   if (status != 0 || (name_list[0].n_value  == 0
@@ -633,6 +667,7 @@ lookup_external_name(char *name, long *location)
 #endif /*! HAVE_DLOPEN */
 }
 
+#endif /* !defined(__hpux) */
 #endif /* !defined(AIX) */
 
 extern char *object_file;   /* specified via a command line argument */
@@ -650,4 +685,3 @@ get_reloc_file()
     return(NULL);
   }
 }
-
